==> thoughtful-2.6.0-py3-none-any.whl: thoughtful/screen_recorder/screen_recorder.py <==
"""Classes for recording screen via ScreenRecorder."""

import atexit
import base64
import datetime
import logging
import os
import queue
import threading
import time
from io import BytesIO
from typing import Any, Optional

from imageio import get_reader, get_writer
from numpy import array, ndarray
from PIL import Image
from SeleniumLibrary.errors import NoOpenBrowser

from thoughtful.screen_recorder.browser_manager import BrowserManager

logger = logging.getLogger(__name__)


class ScreenRecorder:
    """A class to record screen activity of a programmatic browser
    instance and generate a video file.

    The ScreenRecorder class facilitates the recording of screen activity
    from a programmatic browser session and generates a video file of the
    recorded session.

    Hardware requirements:
        The ScreenRecorder requires 1 dedicated thread running in the
        background for screen capturing and a second non-dedicated thread
        for screenshot processing.

    Known limitations:
        - Does not handle screen size changes, only keeps frames of
            the most common size.
        - Does not manage scenarios where a new BrowserManager instance
            must be created (e.g., instance is deleted).

    Future improvements:
        - Prevent calling start_recording if automatically_start_recording
            is True

    Example:
        from RPA.Browser.Selenium import Selenium  # installed separately
        from thoughtful.screen_recorder import ScreenRecorder, BrowserManager

        class YoutubeScraper(ScreenRecorder):
            def __init__(self) -> None:
                self._selenium_instance = Selenium()
                super().__init__(browser_manager=BrowserManager(instance=self._selenium_instance))

        youtube_scraper = YoutubeScraper()
        ... Perform actions here ...
        youtube_scraper.end_recording()
    """

    def __init__(
        self,
        browser_manager: BrowserManager,
        use_real_fps: bool = True,
        automatically_start_recording: bool = True,
        output_folder_path: str = "output",
        output_video_name: Optional[str] = None,
        single_threaded_execution: bool = False,
        # Extremely jank - temp solution for patching ffmpeg dependencies
        video_writer: Any = None,
    ) -> None:
        """Initialize the ScreenRecorder.

        Args:
            browser_manager (BrowserManager): Class providing a common
                interface for interacting with the browser.
            use_real_fps (bool): When recording, we are not able to reliably
                record a constant frame rate due to the time it takes for
                the BrowserManager to take a screenshot. This time is
                dependent on both the underlying hardware and the
                BrowserManager code itself. In order to maximize the number
                of screenshots we are able to take, we limit the logic
                controlling the taking of screenshots. As result, we have
                chosen not to add logic to granularly control the rate at
                which screenshots are taken and in turn have a dynamically
                determined fps that can only be calculated after all frames
                of the video have been taken.

                If `use_real_fps` is True, we will calculate
                self._real_fps after all video frames have been taken
                and adjust the video to utilize self._real_fps. Adjusting
                the video to utilize self._real_fps is a time intensive
                process (about 6 minutes of processing per 60 minutes
                recorded) - to speed up video recording, it is recommended
                to set `use_real_fps` to False.
                If `use_real_fps` is False, we will utilize self._baseline_fps
                which will result in the video being either shorter or longer
                than reality. If self._baseline_fps is lower than
                self._real_fps, the output video will be longer (slower)
                than reality. If self._baseline_fps is higher than
                self._real_fps, the output video will be shorter (faster)
                than reality.
            automatically_start_recording (bool): If True, the recording
                will automatically start as soon as a browser is detected in
                the BrowserManager instance
            output_folder_path (str): Path to folder to store output video file.
            output_video_name (Optional[str]): Name of the output video file.
            single_threaded_execution (bool): The program is being run with
                only one thread (like in Pytest). In this case, we
                cannot have a dedicated screen recording thread - we must
                instead periodically give access back to the scheduler
                to allow operations form other threads to occur.
        """
        self._browser_manager: BrowserManager = browser_manager
        self._use_real_fps = use_real_fps
        self._single_threaded_execution = single_threaded_execution

        if output_video_name is None:
            # If output_video_name is not provided, construct a default name
            # using the child classes name and datetime
            current_datetime = datetime.datetime.now()
            class_name = self.__class__.__name__
            formatted_timestamp_string = current_datetime.strftime("%Y-%m-%d_%H:%M:%S")
            output_video_name = f"{class_name}_{formatted_timestamp_string}_video"
        self._output_filepath_without_extension = os.path.join(
            output_folder_path, output_video_name
        )
        self._file_extension = "mp4"
        self._output_filepath: str = (
            f"{self._output_filepath_without_extension}.{self._file_extension}"
        )

        # Variables used during recording
        self._recording_thread: Optional[threading.Thread] = None
        self._add_frames_to_video_thread: Optional[threading.Thread] = None
        self._stop_capturing_frames_event = threading.Event()
        self._frame_queue = queue.Queue()
        self._recording_length: Optional[int] = None
        self._real_fps: Optional[float] = None
        self._recording_started = False
        self._end_recording_called = False

        # Define settings for video writer and create writer instance
        self._baseline_fps = 1
        self._quality = 5  # must be between 1 and 10
        self._bitrate = "500k"
        self._codec = ""
        self._ffmpeg_log_level = "error"
        self._video_writer = video_writer

        if automatically_start_recording:
            self.start_recording()

        atexit.register(self._check_if_end_recording_called)

    @property
    def is_recording(self) -> bool:
        """Check if the recording is currently in progress."""
        return self._recording_thread is not None and self._recording_thread.is_alive()

    @staticmethod
    def _convert_from_base64_string_to_ndarray(base64_string: str) -> ndarray:
        """Convert base64 string to ndarray."""
        image_data = base64.b64decode(base64_string)
        image = Image.open(BytesIO(image_data))
        image_np_array = array(image)
        return image_np_array

    def _start_capturing_frames(self) -> None:
        """
        Start the frame capturing thread and cleanup after recording has
        been ended.
        """
        start_time = None

        while not self._stop_capturing_frames_event.is_set():
            while (
                self._browser_manager.is_browser_open() is False
                or self._browser_manager.has_page_loaded() is False
            ):
                if self._stop_capturing_frames_event.is_set():
                    raise RuntimeError(
                        "Unable to start capturing frames. Browser not "
                        "opened or window never loaded."
                    )
                time.sleep(0.5)  # Wait until browser opens to start recording

            if start_time is None:
                start_time = time.time()
                logger.info("Beginning video recording.")

            try:
                base64_screenshot = self._browser_manager.get_base64_screenshot()
                self._frame_queue.put(base64_screenshot)
                if self._single_threaded_execution:
                    # If we only have 1 executing thread, we must
                    # periodically return access to the scheduler so
                    # other threads can perform operations such as
                    # signaling to this thread that it should terminate
                    time.sleep(0.5)  # Cede control to other threads
            except NoOpenBrowser:
                logger.info("Skipping video frame due to no browser")

        end_time = time.time()
        self._recording_length = end_time - start_time
        logger.info(f"Recording finished in {self._recording_length:.2f} seconds")

    def _start_adding_frames_to_video(self) -> None:
        logger.info("Starting to add frames to video.")
        frames_captured = 0
        while self._recording_thread.is_alive() or self._frame_queue.empty() is False:
            base64_screenshot = self._frame_queue.get()
            ndarray_screenshot = self._convert_from_base64_string_to_ndarray(
                base64_string=base64_screenshot
            )
            self._video_writer.append_data(ndarray_screenshot)
            frames_captured += 1

            if frames_captured % 100 == 0:
                logger.info(f"{frames_captured} video frames written.")

            # Sleep to allow other threads to get prioritized while waiting
            # additional frames to get added to the queue
            time.sleep(0.5)

        logger.info("All captured frames have been added to the output video.")

        self._real_fps = frames_captured / self._recording_length
        logger.info(
            f"We managed to capture {frames_captured} video frames over "
            f"{self._recording_length:.2f} seconds which translates to "
            f"{self._real_fps:.2f} frames captured per second."
        )

    def _resize_video_to_use_real_frame_rate(self) -> None:
        """
        Update the video to use the real frame rate instead of the baseline
        frame rate.
        """
        logger.info(
            f"Updating frame rate from {self._baseline_fps} fps to"
            f" {self._real_fps} fps."
        )

        try:
            tmp_filename = (
                f"{self._output_filepath_without_extension}_tmp."
                f"{self._file_extension}"
            )
            reader = get_reader(self._output_filepath)
            writer = get_writer(
                tmp_filename,
                fps=self._real_fps,
                codec=self._codec,
                quality=self._quality,
                bitrate=self._bitrate,
                ffmpeg_log_level=self._ffmpeg_log_level,
            )
            for frame in reader:
                writer.append_data(frame)
            writer.close()
            logger.info("Frame rate updated successfully.")

            original_file_size = os.path.getsize(self._output_filepath)
            new_file_size = os.path.getsize(tmp_filename)
            logger.info(
                f"Original video file size: {original_file_size} bytes. "
                f"Updated video file size: {new_file_size} bytes."
            )

            os.remove(self._output_filepath)
            logger.info(f"Deleted old video file ({self._output_filepath}).")

            os.rename(tmp_filename, self._output_filepath)
            logger.info(
                f"Renamed new video file from {tmp_filename} to "
                f"{self._output_filepath}."
            )
        except Exception as e:
            logger.error(f"Unable to update video frame rate: {e}")

    def _check_if_end_recording_called(self) -> None:
        if self._recording_started and not self._end_recording_called:
            raise RuntimeError(
                "'end_recording' not called but main thread has terminated."
            )

    def start_recording(self) -> None:
        """Start the recording thread."""
        self._recording_started = True

        # When instantiating the video_writer instance, the video_writer
        # opens up the video file for writing as part of initialization.
        # Because of this, we wait until the time of recording to initialize
        # the video_writer.
        if self._video_writer is None:
            self._video_writer = get_writer(
                self._output_filepath,
                fps=self._baseline_fps,
                codec=self._codec,
                quality=self._quality,
                bitrate=self._bitrate,
                ffmpeg_log_level=self._ffmpeg_log_level,
            )

        self._recording_thread = threading.Thread(
            target=self._start_capturing_frames,
            # Daemon thread will not block interpreter from exiting
            daemon=True,
        )
        self._recording_thread.start()
        self._add_frames_to_video_thread = threading.Thread(
            target=self._start_adding_frames_to_video,
            # Daemon thread will not block interpreter from exiting
            daemon=True,
        )
        self._add_frames_to_video_thread.start()
        # TODO: Make this sleep statement more robust
        # Wait a moment for thread to start up before returning control to main
        # We want the video recording thread to fully initialize before we
        # allow the main thread to begin performing operations with the
        # BrowserManager
        time.sleep(1)

    def end_recording(self) -> None:
        """Facilitates completion of recording and any cleanup."""
        if not self.is_recording:
            raise RuntimeError(
                "Attempted to end recording, but no recording is in progress."
            )

        self._end_recording_called = True

        self._stop_capturing_frames_event.set()

        logger.info("Blocking until recording thread terminates.")
        self._recording_thread.join()

        logger.info("Blocking until add_frames_to_video thread terminates.")
        self._add_frames_to_video_thread.join()

        logger.info("All supporting threads have been terminated.")

        self._video_writer.close()
        logger.info("Video writing has completed.")

        if self._use_real_fps:
            self._resize_video_to_use_real_frame_rate()

        try:
            # Verify video was successfully created
            file_size = os.path.getsize(self._output_filepath)
            logger.info(
                f"Video recording saved to {self._output_filepath}. "
                f"Video file size: {file_size} bytes."
            )
        except (FileNotFoundError, OSError):
            logger.error(f"Output video file does not exist at {self._output_filepath}")

==> thoughtful-2.7.0-py3-none-any.whl: thoughtful/screen_recorder/screen_recorder.py <==
"""Classes for recording screen via ScreenRecorder."""

import atexit
import base64
import datetime
import logging
import os
import queue
import threading
import time
from io import BytesIO
from typing import Any, Optional

from imageio import get_reader, get_writer
from numpy import array, ndarray
from PIL import Image
from selenium.common.exceptions import InvalidSessionIdException
from SeleniumLibrary.errors import NoOpenBrowser

from thoughtful.screen_recorder.browser_manager import BrowserManager

logger = logging.getLogger(__name__)


class ScreenRecorder:
    """A class to record screen activity of a programmatic browser
    instance and generate a video file.

    The ScreenRecorder class facilitates the recording of screen activity
    from a programmatic browser session and generates a video file of the
    recorded session.

    Hardware requirements:
        The ScreenRecorder requires 1 dedicated thread running in the
        background for screen capturing and a second non-dedicated thread
        for screenshot processing.

    Known limitations:
        - Does not handle screen size changes, only keeps frames of
            the most common size.
        - Does not manage scenarios where a new BrowserManager instance
            must be created (e.g., instance is deleted).

    Future improvements:
        - Prevent calling start_recording if automatically_start_recording
            is True

    Example:
        from RPA.Browser.Selenium import Selenium  # installed separately
        from thoughtful.screen_recorder import ScreenRecorder, BrowserManager

        class YoutubeScraper(ScreenRecorder):
            def __init__(self) -> None:
                self._selenium_instance = Selenium()
                super().__init__(browser_manager=BrowserManager(instance=self._selenium_instance))

        youtube_scraper = YoutubeScraper()
        ... Perform actions here ...
        youtube_scraper.end_recording()
    """

    def __init__(
        self,
        browser_manager: BrowserManager,
        use_real_fps: bool = True,
        automatically_start_recording: bool = True,
        output_folder_path: str = "output",
        output_video_name: Optional[str] = None,
        single_host_thread: bool = True,
        # Extremely jank - temp solution for patching ffmpeg dependencies
        video_writer: Any = None,
    ) -> None:
        """Initialize the ScreenRecorder.

        Args:
            browser_manager (BrowserManager): Class providing a common
                interface for interacting with the browser.
            use_real_fps (bool): When recording, we are not able to reliably
                record a constant frame rate due to the time it takes for
                the BrowserManager to take a screenshot. This time is
                dependent on both the underlying hardware and the
                BrowserManager code itself. In order to maximize the number
                of screenshots we are able to take, we limit the logic
                controlling the taking of screenshots. As result, we have
                chosen not to add logic to granularly control the rate at
                which screenshots are taken and in turn have a dynamically
                determined fps that can only be calculated after all frames
                of the video have been taken.

                If `use_real_fps` is True, we will calculate
                self._real_fps after all video frames have been taken
                and adjust the video to utilize self._real_fps. Adjusting
                the video to utilize self._real_fps is a time intensive
                process (about 6 minutes of processing per 60 minutes
                recorded) - to speed up video recording, it is recommended
                to set `use_real_fps` to False.
                If `use_real_fps` is False, we will utilize self._baseline_fps
                which will result in the video being either shorter or longer
                than reality. If self._baseline_fps is lower than
                self._real_fps, the output video will be longer (slower)
                than reality. If self._baseline_fps is higher than
                self._real_fps, the output video will be shorter (faster)
                than reality.
            automatically_start_recording (bool): If True, the recording
                will automatically start as soon as a browser is detected in
                the BrowserManager instance
            output_folder_path (str): Path to folder to store output video file.
            output_video_name (Optional[str]): Name of the output video file.
            single_host_thread (bool): The program is being run with
                only one thread/vCPU. In this case, we cannot have a dedicated
                screen recording thread - we must instead periodically give
                access back to the scheduler to allow operations form other
                threads to occur.
        """
        self._browser_manager: BrowserManager = browser_manager
        self._use_real_fps = use_real_fps
        self._single_host_thread = single_host_thread

        if output_video_name is None:
            # If output_video_name is not provided, construct a default name
            # using the child classes name and datetime
            current_datetime = datetime.datetime.now()
            class_name = self.__class__.__name__
            formatted_timestamp_string = current_datetime.strftime("%Y-%m-%d_%H:%M:%S")
            output_video_name = f"{class_name}_{formatted_timestamp_string}_video"
        self._output_filepath_without_extension = os.path.join(
            output_folder_path, output_video_name
        )
        self._file_extension = "mp4"
        self._output_filepath: str = (
            f"{self._output_filepath_without_extension}.{self._file_extension}"
        )

        # Variables used during recording
        self._recording_thread: Optional[threading.Thread] = None
        self._add_frames_to_video_thread: Optional[threading.Thread] = None
        self._stop_capturing_frames_event = threading.Event()
        self._frame_queue = queue.Queue()
        self._recording_length: Optional[int] = None
        self._real_fps: Optional[float] = None
        self._recording_started = False
        self._end_recording_called = False

        # Define settings for video writer and create writer instance
        self._baseline_fps = 1
        self._quality = 5  # must be between 1 and 10
        self._bitrate = "500k"
        self._codec = ""
        self._ffmpeg_log_level = "error"
        self._video_writer = video_writer

        if automatically_start_recording:
            self.start_recording()

        atexit.register(self._check_if_end_recording_called)

    @property
    def is_recording(self) -> bool:
        """Check if the recording is currently in progress."""
        return self._recording_thread is not None and self._recording_thread.is_alive()

    @staticmethod
    def _convert_from_base64_string_to_ndarray(base64_string: str) -> ndarray:
        """Convert base64 string to ndarray."""
        image_data = base64.b64decode(base64_string)
        image = Image.open(BytesIO(image_data))
        image_np_array = array(image)
        return image_np_array

    def _start_capturing_frames(self) -> None:
        """
        Start the frame capturing thread and cleanup after recording has
        been ended.
        """
        start_time = None

        while not self._stop_capturing_frames_event.is_set():
            while (
                self._browser_manager.is_browser_open() is False
                or self._browser_manager.has_page_loaded() is False
            ):
                if self._stop_capturing_frames_event.is_set():
                    raise RuntimeError(
                        "Unable to start capturing frames. Browser not "
                        "opened or window never loaded."
                    )
                time.sleep(0.5)  # Wait until browser opens to start recording

            if start_time is None:
                start_time = time.time()
                logger.info("Beginning video recording.")

            try:
                base64_screenshot = self._browser_manager.get_base64_screenshot()
                self._frame_queue.put(base64_screenshot)
            except NoOpenBrowser:
                logger.warning("Skipping video frame due to no browser")
            except InvalidSessionIdException:
                logger.warning(
                    "Skipping video frame due to invalid session id. This "
                    "likely means the session crashed or was deleted."
                )
            if self._single_host_thread:
                # If we only have 1 executing thread/vCPU, we must
                # periodically return access to the scheduler so
                # other threads can perform operations such as
                # signaling to this thread that it should terminate
                time.sleep(0.3)  # Cede control to other threads

        end_time = time.time()
        self._recording_length = end_time - start_time
        logger.info(f"Recording finished in {self._recording_length:.2f} seconds")

    def _start_adding_frames_to_video(self) -> None:
        logger.info("Starting to add frames to video.")
        frames_captured = 0
        while self._recording_thread.is_alive() or self._frame_queue.empty() is False:
            base64_screenshot = self._frame_queue.get()
            ndarray_screenshot = self._convert_from_base64_string_to_ndarray(
                base64_string=base64_screenshot
            )
            self._video_writer.append_data(ndarray_screenshot)
            frames_captured += 1

            if frames_captured % 100 == 0:
                logger.info(f"{frames_captured} video frames written.")

            # Sleep to allow other threads to get prioritized while waiting
            # additional frames to get added to the queue
            time.sleep(0.5)

        logger.info("All captured frames have been added to the output video.")

        self._real_fps = frames_captured / self._recording_length
        logger.info(
            f"We managed to capture {frames_captured} video frames over "
            f"{self._recording_length:.2f} seconds which translates to "
            f"{self._real_fps:.2f} frames captured per second."
        )

    def _resize_video_to_use_real_frame_rate(self) -> None:
        """
        Update the video to use the real frame rate instead of the baseline
        frame rate.
        """
        logger.info(
            f"Updating frame rate from {self._baseline_fps} fps to"
            f" {self._real_fps} fps."
        )

        try:
            tmp_filename = (
                f"{self._output_filepath_without_extension}_tmp."
                f"{self._file_extension}"
            )
            reader = get_reader(self._output_filepath)
            writer = get_writer(
                tmp_filename,
                fps=self._real_fps,
                codec=self._codec,
                quality=self._quality,
                bitrate=self._bitrate,
                ffmpeg_log_level=self._ffmpeg_log_level,
            )
            for frame in reader:
                writer.append_data(frame)
            writer.close()
            logger.info("Frame rate updated successfully.")

            original_file_size = os.path.getsize(self._output_filepath)
            new_file_size = os.path.getsize(tmp_filename)
            logger.info(
                f"Original video file size: {original_file_size} bytes. "
                f"Updated video file size: {new_file_size} bytes."
            )

            # Rename old file instead of deleting for now. We want to
            # delete the file as the very last step in the process incase
            # there are any failures in the intermediary steps.
            tmp_filename_old_video = (
                f"{self._output_filepath_without_extension}_tmp_old."
                f"{self._file_extension}"
            )
            os.rename(self._output_filepath, tmp_filename_old_video)
            logger.info(
                f"Renamed old video file from {self._output_filepath} to "
                f"{tmp_filename_old_video}."
            )

            os.rename(tmp_filename, self._output_filepath)
            logger.info(
                f"Renamed new video file from {tmp_filename} to "
                f"{self._output_filepath}."
            )

            os.remove(tmp_filename_old_video)
            logger.info(f"Deleted old video file ({tmp_filename_old_video}).")

        except Exception as e:
            logger.error(f"Unable to update video frame rate: {e}")

    def _check_if_end_recording_called(self) -> None:
        if self._recording_started and not self._end_recording_called:
            raise RuntimeError(
                "'end_recording' not called but main thread has terminated."
            )

    def start_recording(self) -> None:
        """Start the recording thread."""
        self._recording_started = True

        # When instantiating the video_writer instance, the video_writer
        # opens up the video file for writing as part of initialization.
        # Because of this, we wait until the time of recording to initialize
        # the video_writer.
        if self._video_writer is None:
            self._video_writer = get_writer(
                self._output_filepath,
                fps=self._baseline_fps,
                codec=self._codec,
                quality=self._quality,
                bitrate=self._bitrate,
                ffmpeg_log_level=self._ffmpeg_log_level,
            )

        self._recording_thread = threading.Thread(
            target=self._start_capturing_frames,
            # Daemon thread will not block interpreter from exiting
            daemon=True,
        )
        self._recording_thread.start()
        self._add_frames_to_video_thread = threading.Thread(
            target=self._start_adding_frames_to_video,
            # Daemon thread will not block interpreter from exiting
            daemon=True,
        )
        self._add_frames_to_video_thread.start()
        # TODO: Make this sleep statement more robust
        # Wait a moment for thread to start up before returning control to main
        # We want the video recording thread to fully initialize before we
        # allow the main thread to begin performing operations with the
        # BrowserManager
        time.sleep(1)

    def end_recording(self) -> None:
        """Facilitates completion of recording and any cleanup."""
        if not self.is_recording:
            raise RuntimeError(
                "Attempted to end recording, but no recording is in progress."
            )

        self._end_recording_called = True

        self._stop_capturing_frames_event.set()

        logger.info("Blocking until recording thread terminates.")
        self._recording_thread.join()

        logger.info("Blocking until add_frames_to_video thread terminates.")
        self._add_frames_to_video_thread.join()

        logger.info("All supporting threads have been terminated.")

        self._video_writer.close()
        logger.info("Video writing has completed.")

        if self._use_real_fps:
            try:
                self._resize_video_to_use_real_frame_rate()
            except Exception as e:
                logger.error(f"Unable to resize video frame rate: {e}")

        try:
            # Verify video was successfully created
            file_size = os.path.getsize(self._output_filepath)
            logger.info(
                f"Video recording saved to {self._output_filepath}. "
                f"Video file size: {file_size} bytes."
            )
        except (FileNotFoundError, OSError):
            logger.error(f"Output video file does not exist at {self._output_filepath}")
